from fanery.terms import date, datetime
from dateutil import relativedelta

class relativedelta(relativedelta.relativedelta):

    def __init__(self, *args, **argd):
        self.workdays = argd.pop('workdays', 0)
        if self.workdays:
            assert self.workdays > 0, 'workdays must be int > 0'
            self.calendar = cal = argd.pop('calendar', None)
            assert cal, 'workdays requires calendar'
        super(relativedelta, self).__init__(*args, **argd)

    def __add__(self, other):
        if self.workdays and isinstance(other, date):
            from dateutil.rrule import rrule, DAILY
            if not isinstance(other, datetime):
                other = datetime.fromordinal(other.toordinal())
            years = set()
            holidays = set()
            cal = self.calendar
            days = self.workdays
            rrule = rrule(DAILY, byweekday = cal.WORKDAY, dtstart = other)
            for day in rrule:
                if day.year not in years:
                    holidays.update(d.date() for d in cal.holidays(day.year))
                    years.add(day.year)
                if day.date() in holidays:
                    continue
                elif days > 0:
                    days -= 1
                else:
                    return day
        else:
            return super(relativedelta, self).__add__(other)

def easter(year = None):
    # from https://en.wikipedia.org/wiki/Computus
    year = year or date.today().year
    a = year % 19
    b = year >> 2
    c = b // 25 + 1
    d = (c * 3) >> 2
    e = ((a * 19) - ((c * 8 + 5) // 25) + d + 15) % 30
    e += (29578 - a - e * 32) >> 10
    e -= ((year % 7) + b - d + e + 2) % 7
    d = e >> 5
    day = e - d * 31
    month = d + 3
    return datetime(year, month, day)




from fanery.librand import *
from fanery.terms import *

def test_hict():
    d = hict(n = randnum())
    assert d.n is d['n']
    assert isinstance(d.n, (int, long, Decimal))
    assert is_dict(d.d)
    assert is_dict(d.d.d)
    assert is_dict(d.d.d.d)
    s = ('2012-01-01', '1.0', '1')
    d.s = s
    assert is_date(d.s[0]), d
    assert is_number(d.s[1]), d
    assert is_number(d.s[2]), d
    d.setdefault('n', '2000-01-01')
    assert is_number(d.n)
    d.update(n = '2000-01-01', n02 = 'Nov 02')
    assert is_date(d.n)
    assert d.n is d['n']
    assert is_date(d.n02), d
    assert d.n02.month is 11 and d.n02.day is 2
    assert d.update(e = 1.0).update(f = 1) is d
    assert d.e is d.f is 1



from fanery import service, authorize, store
from fanery.terms import (
    hict, get_timestamp, is_string, is_string,
    parse_term, parse_json, to_json,
)
from fanery.crypto import gen_token, digest, encrypt, decrypt

class PyClient(object):

    def __init__(self):
        self.session = hict(token = None, sid = None)

    def call(self, urlpath, *args, **argd):
        return service.consume(urlpath, *args, **argd)

    def safe_call(self, urlpath, *args, **argd):
        session = self.session
        sid = session.sid
        key = session.token
        params = argd.pop('params', dict())
        enc_call = encrypt(key, [urlpath, args, argd])
        params.update(sid = sid, enc_call = enc_call)
        enc_ret = self.call('safe_call', **params)
        ret = decrypt(key, enc_ret)
        session.token = ret.token
        return parse_term(ret.data)

    def login(self, username, password, **argd):
        password = digest(password)
        tstamp = get_timestamp()
        nonce = gen_token()
        key = digest('%s%s%f' % (nonce, password, tstamp))
        msg = digest('%s%s%s' % (username, nonce, password))
        ret = self.call('login', username = username,
                                 nonce = nonce,
                                 tstamp = tstamp,
                                 enc_token = encrypt(key, msg),
                                 **argd)
        self.session.update(decrypt(key, ret))
        return True

    def logout(self):
        try:
            return self.safe_call('logout')
        finally:
            self.session.update(token = None, sid = None)

from fanery.wsgi import JSON_KEY, handler
from webob import Request

class WSGIClient(PyClient):

    def send(self, urlpath, **argd):
        req = Request.blank(urlpath, POST = {JSON_KEY: to_json(argd)})
        return req.get_response(handler)

    def call(self, urlpath, *args, **argd):
        path = '/'.join(['', urlpath, '/'.join(str(i) for i in args)])
        res = self.send(path, **argd)
        if res.status_code == 200:
            try:
                return parse_json(res.body)
            except:
                return res.body
        elif res.status_code == 400:
            try:
                error = parse_json(res.body)
                etype = getattr(store, error.type, None) or \
                        getattr(service, error.type, None) or \
                        getattr(authorize, error.type, None) or \
                        getattr(__builtins__, error.type, Exception)
                reason = error.reason
            except:
                raise Exception(res.body)
            if issubclass(etype, Exception):
                raise etype(*reason)
            else:
                raise Exception(*reason)
        else:
            raise Exception(res.body)

class HTTPClient(WSGIClient):

    def send(self, urlpath, **argd):
        raise NotImplementedError



from os.path import isdir as _isdir, isfile as _isfile
from dateutil.parser import parse as _parse_date
from unicodedata import normalize as _normalize

from decimal import Decimal
from datetime import date, datetime
from time import time as get_timestamp
from uuid import UUID

from inspect import (
    ismodule as is_module,
    isclass as is_class,
    ismethod as is_method,
    isfunction as is_function,
    isgenerator as is_generator,
)

def is_subclass(term, *args, **argd):
    try:
        return issubclass(term, *args, **argd)
    except:
        return False

def is_lambda(term):
    try:
        return is_function(term) and term.__name__ == is_lambda.__lambda_name__
    except:
        return False
is_lambda.__lambda_name__ = (lambda: None).__name__

def is_file(term):
    try:
        return _isfile(term)
    except:
        return False

def is_dir(term):
    try:
        return _isdir(term)
    except:
        return False

def is_boolean(term):
    return isinstance(term, bool)

def is_uuid(term):
    return isinstance(term, UUID)

def is_date(term):
    return isinstance(term, (date, datetime))

def is_dict(term):
    return isinstance(term, dict) or type(term) == dict

def is_sequence(term):
    return hasattr(term, '__iter__') \
            and not isinstance(term, basestring) \
            and not is_dict(term)

def is_record(term, model = None):
    return isinstance(term, store.Record) \
            and (model is None or term._model == model)

def is_string(term):
    return isinstance(term, basestring)

def is_number(term):
    return isinstance(term, (int, long, float, Decimal))

_record_dump_keys = set(('_model', '_uuid', '_txn', '_vsn', '_time', '_fields'))
def is_record_dump(term):
    return is_dict(term) and set(term.keys()) == _record_dump_keys \
            and term['_model'] in store.models and is_uuid(term['_uuid']) \
            and is_number(term['_vsn']) and is_dict(term['_fields'])

def is_record_ref(term, model = None):
    try:
        rtype, uuid = term
        return is_uuid(uuid) and is_string(rtype) \
                and (model == rtype if model else True) \
                and rtype in store.models
    except:
        return False

def parse_number(term):
    dec = Decimal(term if isinstance(term, basestring) else '%f' % term)
    num = int(dec)
    return num if num == dec else dec.normalize()

def parse_date(term):
    dt = _parse_date(term)
    d = dt.date()
    return d if d == dt else d

def parse_term(term, parse_float = parse_number, seen = None):
    """Transform term to python object."""
    # cache parsed terms (prevent recursion)
    try:
        tid = hash(term)
    except:
        tid = id(term)
    if seen is None:
        seen = dict()
    elif tid in seen:
        return seen[tid]

    if isinstance(term, (int, long, Decimal, hict, UUID, store.Record)) \
       or is_date(term) or is_file(term):
        pass
    elif isinstance(term, float):
        term = parse_float(term)
    elif term and is_string(term):
        try:
            term = parse_float(term)
        except:
            try:
                term = UUID(term)
            except:
                if len(term) > 2:
                    try:
                        term = parse_date(term)
                    except:
                        pass
    elif is_sequence(term):
        term = type(term)(parse_term(i, parse_float, seen) for i in term)
        if is_record_ref(term):
            with store() as db:
                term = db.get(*term)
    elif is_dict(term):
        term = hict()._fill((parse_term(k, parse_float, seen),
                             parse_term(v, parse_float, seen))
                            for k, v in term.items())
        if is_record_dump(term):
            term = store.new(term._model,
                             term._uuid,
                             term._txn,
                             term._vsn,
                             term._time,
                             term._fields)

    try:
        seen[hash(term)] = term
    except:
        seen[id(term)] = term
    seen[tid] = term
    return term

def to_simple(term, seen = None, levels = 1):
    """Transform python object to simple serializable term."""
    # cache transformation (prevent recursion)
    try:
        tid = hash(term)
    except:
        tid = id(term)
    if seen is None:
        seen = dict()
    elif tid in seen:
        return seen[tid]

    if term is None or isinstance(term, (int, long, basestring)):
        pass
    elif isinstance(term, float):
        term = to_simple(parse_number(term), seen, levels)
    elif isinstance(term, store.Record):
        if levels > 0:
            term = to_simple(term.dump(), seen, levels-1)
        else:
            term = (term._model, str(term._uuid))
    elif is_dict(term):
        term = dict((to_simple(k, seen, levels),
                     to_simple(v, seen, levels))
                    for k,v in term.items())
    elif is_sequence(term) or is_generator(term):
        term = list(to_simple(i, seen, levels) for i in term)
    else:
        term = str(term)

    try:
        seen[hash(term)] = term
    except:
        seen[id(term)] = term
    seen[tid] = term
    return term

try:
    from BeautifulSoup import BeautifulSoup as _to_unicode
    def to_str(term, encoding = 'utf8', fromEncoding = None):
        if isinstance(term, unicode):
            return term.encode('utf8')
        else:
            text = term if isinstance(term, str) else str(to_simple(term))
            soup = _to_unicode(text, fromEncoding = fromEncoding)
            return unicode(soup).encode(encoding)
except ImportError:
    def to_str(term, encoding = 'utf8', fromEncoding = None):
        if isinstance(term, unicode):
            return term.encode('utf8')
        elif fromEncoding:
            return text.decode(fromEncoding).encode(encoding)
        else:
            text = term if isinstance(term, str) else str(to_simple(term))
            for e in ('utf8', 'latin1', 'windows-1250', 'windows-1252'):
                try:
                    return text.decode(e).encode(encoding)
                except UnicodeDecodeError, err:
                    continue
            raise err

def to_unicode(text, encoding = 'utf8', fromEncoding = None):
    return to_str(text, encoding, fromEncoding).decode(encoding)

def to_ascii(text, errors = 'ignore', form = 'NFKD'):
    return _normalize(form, text).encode('ascii', errors)

try:
    from cPickle import dumps as _pickle, loads as _unpickle
except ImportError:
    from Pickle import dumps as _pickle, loads as _unpickle
finally:
    def pickle(term):
        return _pickle(to_simple(term))

    def unpickle(term):
        return parse_term(_unpickle(term))

try:
    from yajl import dumps as _to_json, loads as _from_json
except ImportError:
    try:
        from simplejson import dumps as _to_json, loads as _from_json
    except ImportError:
        from json import dumps as _to_json, loads as _from_json
finally:
    def to_json(term):
        return _to_json(to_simple(term))

    def parse_json(term):
        return parse_term(_from_json(term))


class hict(dict):
    """Hierarchycal dotted dictionary with auto-parsing."""

    def __init__(self, *args, **argd):
        if args or argd:
            self.update(*args, **argd)

    def __contains__(self, key):
        return dict.__contains__(self, parse_term(key))

    def __getitem__(self, key):
        k = parse_term(key)
        if not dict.__contains__(self, k):
            return dict.setdefault(self, k, hict())
        else:
            return dict.__getitem__(self, k)

    def __setitem__(self, key, val):
        dict.__setitem__(self, parse_term(key), parse_term(val))

    __getattr__ = __getitem__
    __setattr__ = __setitem__
    __delattr__ = dict.__delitem__

    def _fill(self, *args, **argd):
        if args or argd:
            for k, v in dict(*args, **argd).items():
                dict.__setitem__(self, k, v)
        return self

    def setdefault(self, key, term):
        k = parse_term(key)
        if not dict.__contains__(self, k):
            dict.__setitem__(self, k, parse_term(term))

    def update(self, *args, **argd):
        if args or argd:
            for k, v in dict(*args, **argd).items():
                self[k] = v
        return self

    def merge(self, *args, **argd):
        if args or argd:
            for k, v in dict(*args, **argd).items():
                self.setdefault(k, v)
        return self

    def difference(self, dct):
        return hict((k, v) for k, v in self.items() if k not in dct)

    def union(self, dct):
        return hict(self).update(dct)

    def remove(self, term):
        for key in self.find_keys(term):
            del self[key]
        return self

    def find_keys(self, *args):
        vals = (parse_term(arg) for arg in args)
        for k, v in self.items():
            if v in vals:
                yield k



from rawes.elastic_exception import ElasticException
from rawes import Elastic

from fanery.terms import parse_term, to_simple, to_json, is_dict
from fanery import store, service

default_params = dict()

class Model(object):

    def __init__(self, name, es):
        self.name = name
        self.es = es.es

    @property
    def _index(self):
        try:
            return self.__index
        except AttributeError:
            index = service.state.domain or 'local'
            es = self.es
            try:
                es.get('%s/_status' % index)
            except ElasticException:
                es.put(index)
            finally:
                self.__index = idx = (es, '%s/%s' % (index, self.name))
                return idx

    def get(self, uuid):
        es, path = self._index
        rs = es.get('%s/%s' % (path, uuid))
        if rs['exists']:
            return parse_term(rs['_source'])
        else:
            return None

    def fetch(self, **argd):
        record = None
        for idx, record in enumerate(self.select(**argd)):
            if idx > 0:
                raise store.MultipleRecordsFound()
        return record

    def select(self, **argd):
        fields = dict((k, argd.pop(k)) for k,v in argd.items() if not is_dict(v))
        if fields:
            query = ' AND '.join('%s:%s' % (f, to_json(v)) if v is not None \
                                 else '_missing_:%s' % f for f,v in fields.items())
            if query:
                argd['data'] = {'query': {'query_string': {'query': query}}}
        es, path = self._index
        for hit in es.get('%s/_search' % path, **argd)['hits']['hits']:
            yield parse_term(hit['_source'])

    def insert(self, record):
        es, path = self._index
        es.put('%s/%s' % (path, record._uuid),
               data = to_simple(record.dump()),
               params = default_params)

    def update(self, record):
        es, path = self._index
        es.put('%s/%s?version=%d' % (path, record._uuid, record._vsn-1),
               data = to_simple(record.dump()),
               params = default_params)

    def delete(self, record):
        es, path = self._index
        es.delete('%s/%s?version=%d' % (path, record._uuid, abs(record._vsn)),
                  params = default_params)

    def clear(self):
        es, path = self._index
        try:
            es.delete(path, params = default_params)
        except ElasticException:
            pass

class ElasticSearch(object):

    def __init__(self, *args, **argd):
        argd.setdefault('except_on_error', True)
        self.es = Elastic(*args, **argd)

    def begin(self, txn):
        pass

    def prepare_commit(self, txn):
        pass

    def commit(self, txn):
        pass

    def rollback(self, txn):
        pass

    def model(self, name):
        return Model(name, self)



from fanery.filelib import splitext, fullpath, joinpath, isdir, isfile
from fanery.terms import (
    hict, parse_term, get_timestamp, is_string,
    to_simple, to_json, to_str, pickle,
)
from fanery.librand import randuuid

from inspect import currentframe
from functools import wraps
import sys, cgitb, logging

class service(object):

    production = True
    profile = False

    class UnknownFormatter(AssertionError):
        pass

    class SslRequired(AssertionError):
        pass

    class UnsafeCall(AssertionError):
        pass

    class NotFound(AssertionError):
        pass

    def __init__(self):
        self.registry = hict()
        self.batch_jobs = dict()

    def __call__(self, path = None, safe = True, ssl = True, cache = False,
                 static = False, force_download = False, auto_parse = True,
                 log_exc = False, one_time_pad = True, check_outfmt = False,
                 **out):
        '''Register a new service.

        path           - urlpath prefix
        safe           - boolean that enforce service consumption via safe_call
        ssl            - boolean that enforce service consumption over ssl
        cache          - boolean that set/unset response header settings for
                         client side caching. NOTE: no server side caching
        static         - boolean that tell if results are filesystem elements
        force_download - boolean that tell if results are attachments
        auto_parse     - boolean that say if call *args/**kwargs must be parsed
                         terms parsing is performed by fanery.terms.parse_term
        one_time_pad   - boolean that say if crypto token must be unique, which
                         means that a new token is generated on every safe call
        log_exc        - boolean that say if exception must be logged
        check_outfmt   - boolean that say if result formatter must be defined
        **out          - result formatters where key represents mime extension
                         and value is the corresponding callable
                       : default formatters if check_outfmt is not True
                            txt, raw ==> fanery.terms.to_str
                            json     ==> fanery.terms.to_json
                            bin      ==> fanery.terms.pickle

        Example: define a service exposed as /gen_report that accept
                 authenticated safe call over ssl only for users with
                 "board-member" role and return pdf/ods formatted reports.

            from fanery import service, authorize

            def report_pdf(data):
                transform data into pdf file
                return pdf file object or pdf file path

            def report_ods(data):
                transform data into ods file
                return ods file object or ods file path

            @authorize('board-member')
            @service(static = True, ods = report_ods, pdf = report_pdf)
            def gen_report(*args, **argd):
                organize report data based on call parameters
                return data
        '''
        from fanery.profiler import profile

        def decorator(func):

            @wraps(func)
            def wrapper(*args, **argd):
                state = self.state
                if self.production:
                    if ssl is True and state.ssl is not True:
                        raise service.SslRequired(wrapper.service)
                    elif safe is True and state.safe is not True:
                        raise service.UnsafeCall(wrapper.service)

                if auto_parse is True:
                    args = parse_term(args)
                    argd = parse_term(argd)

                try:
                    if self.profile:
                        print '>>> profiling: %s:%s\n' % (func.__module__, func.__name__)
                        return profile(func)(*args, **argd)
                    else:
                        return func(*args, **argd)
                except Exception, e:
                    if log_exc is True or not isinstance(e, AssertionError):
                        logging.error(cgitb.text(sys.exc_info()))
                    raise

            if check_outfmt is not True:
                out.setdefault('json', to_json)
                out.setdefault('raw', to_str)
                out.setdefault('txt', to_str)
                out.setdefault('bin', pickle)

            wrapper.service = hict(
                ssl = ssl,
                safe = safe,
                cache = cache,
                static = static,
                force_download = force_download,
                one_time_pad = one_time_pad,
                log_exc = log_exc,
                auto_parse = auto_parse,
                route = (path or func.__name__).strip('/'),
                output = dict((k, v) for k, v in out.items() if callable(v)),
                check_outfmt = check_outfmt,
            )

            reg = self.registry
            for part in wrapper.service.route.split('/'):
                reg = reg[part]
            else:
                reg[None] = wrapper

            return wrapper

        return decorator

    def lookup(self, urlpath):
        '''Find a registered service by urlpath.'''
        path, ext = splitext(urlpath.strip('/'))
        registry = reg = self.registry

        parts = path.split('/')
        for idx, part in enumerate(parts):
            if part not in reg:
                if reg is registry:
                    continue # skip context prefix
                else:
                    args = parts[idx:]
                    break
            else:
                reg = reg[part]
        else:
            args = parts if reg is registry else list()
            reg = reg.get('', reg)

        func = reg.get(None, None)
        if not callable(func):
            raise self.NotFound()

        service = func.service
        if ext:
            output = service.output.get(ext[1:], None)
            if args and service.static is True:
                args[-1] = args[-1] + ext
        else:
            output = None

        return func, args, ext, output

    def _consume(self, urlpath, *args, **argd):
        '''Consume a registered service by urlpath.'''
        func, urlpath_args, ext, output = self.lookup(urlpath)
        if func.service.check_outfmt is True and not (ext and output):
            raise service.UnknownFormatter(ext)
        ret = func(*(urlpath_args or args), **argd)
        return func.service, ext, output, output(ret) if output else ret

    def consume(self, urlpath, *args, **argd):
        '''Consume a registered service by urlpath.'''
        return self._consume(urlpath, *args, **argd)[-1]

    @property
    def state(self, _key_ = '_fnry_state_'):
        '''Like thread local session, just without thread.'''
        current = currentframe()
        if _key_ in current.f_locals:
            return current.f_locals[_key_]

        frame = caller = current.f_back
        while frame:
            if _key_ in frame.f_locals:
                state = frame.f_locals[_key_]
                break
            frame = frame.f_back
        else:
            state = hict(domain = 'localhost', ssl = False,
                         safe = False, user = None, data = hict())

        return caller.f_locals.setdefault(_key_,
                current.f_locals.setdefault(_key_, state))

    def static(self, path, root, **extra):
        '''Register a new service that allow to retrieve static contents.

        path - urlpath prefix
        root - folder confinement (kind of chroot).

        Example:

            static('/public', '/var/www/public/', safe = False, ssl = False)
                -> no ssl and no authentication required

            static('/reports', '/var/www/reports/', safe = False)
                -> no authentication required but ssl required

            static('/private', '/var/www/private/')
                -> both ssl and authentication required (safe_call over ssl)
        '''
        root = joinpath(fullpath(root), '').rstrip('/')
        assert isdir(root), 'No such directory "%s"' % root

        index = extra.pop('index', 'index.html')
        extra.setdefault('safe', False)
        extra.setdefault('ssl', False)

        @self(path, static = True, auto_parse = False, **extra)
        def static(*args, **argd):
            dompath = fullpath(joinpath(root, self.state.domain, *args))
            if dompath.startswith(root):
                if isdir(dompath):
                    return joinpath(dompath, index)
                elif isfile(dompath):
                    return dompath
            urlpath = fullpath(joinpath(root, *args))
            if urlpath.startswith(root):
                if isdir(urlpath):
                    return joinpath(urlpath, index)
                elif isfile(urlpath):
                    return urlpath
            raise self.NotFound()

    def batch_job(self, job_id):
        job = self.batch_jobs.get(job_id, None)
        if job and not job.process.is_alive():
            del self.batch_jobs[job_id]
        return job

    def _job(self, func, queue, args, argd):
        from fanery import service
        service.state.queue = queue
        return func(*args, **argd)

    def _watcher(self, job, timeout, cleanup):
        try:
            job.start()
            job.join(timeout)
            if job.is_alive():
                job.terminate()
        finally:
            if callable(cleanup):
                cleanup()

    def batch(self, timeout = None, queue = None, cleanup = None):
        '''Make decorated function run as batch process (not thread).

        Running batch process can be queried for working status at any time:

        @service.batch(timeout = 5)
        def batch_job(*args, **argd):
            # this will only perform half of it's work
            from time import sleep
            # queue by default is a synchronized multiprocessing.Manager.list
            queue = service.state.queue
            for i in range(1,11):
                sleep(1)
                queue.append(10.0/i)

        @service()
        def start_batch_job(*args, **argd):
            # remote client call this and get batch job id
            do something
            job = batch_job() # non-blocking
            do something
            return job.id

        @service()
        def job_progress(job_id):
            # later remote client can query batch job progress by id
            job = service.batch_job(job_id)
            assert job, 'not-found'
            return dict(alive = job.process.is_alive(),
                        progress = job.queue[-1] if len(job.queue) else 0)
        '''
        from multiprocessing import Process, Manager
        from threading import Thread

        def decorator(f):

            @wraps(f)
            def wrapper(*args, **argd):
                job_id, q = randuuid(), queue or Manager().list()
                job = self.batch_jobs[job_id] = self.state.jobs[job_id]
                p = Process(target = self._job, args = (f, q, args, argd))
                Thread(target = self._watcher, args = (p, timeout, cleanup)).start()
                return job.update(id = job_id, process = p, queue = q).id

            return wrapper

        return decorator

    def test(self, setUp = None, tearDown = None):
        self.production = False

        def decorator(func):

            @wraps(func)
            def wrapper(*args, **argd):
                state = self.state
                if callable(setUp):
                    setUp()
                if callable(tearDown):
                    try:
                        func(*args, **argd)
                    finally:
                        tearDown()
                else:
                    func(*args, **argd)

            return wrapper

        return decorator



ISO2_CODES = set()

def load_modules():
    from fanery.filelib import filedir, listdir, isdir, isfile, joinpath, splitext
    for name in listdir(filedir(__file__)):
        iso2 = splitext(name)[0] if isfile(name) else (name if isdir(name) else None)
        if iso2 and len(iso2) == 2 and iso2.isupper():
            ISO2_CODES.add(iso2)

load_modules()



from os.path import abspath, realpath, normpath, curdir, splitext
from os.path import isfile, isdir, exists
from os.path import basename as filename
from os.path import getsize as filesize
from os.path import dirname as filedir
from os.path import join as joinpath
from os import listdir

def files(*args, **argd):
    return dict((f.name, f) for f in args + argd.values() if isfile(f))

def fullpath(path):
    return normpath(realpath(abspath(path)))



from fanery.filelib import joinpath, files, tempdir, savefile, filename, filesize
from fanery import service, static, store, authorize, crypto, dbms

store.register(dbms.DummyMemoryDB(), store.models_names)

Session = authorize.Session.model_name
Profile = authorize.Profile.model_name

@service(safe = False, _log_exc = False)
def login(username, nonce, tstamp, enc_token, force = False, **argd):
    with store() as db:
        user = db.fetch(Profile, login = username, fail = None, **argd)
        assert user is not None, 'username'
        key = '%s%s%s' % (nonce, user.password, tstamp)
        msg = crypto.digest('%s%s%s' % (user.login, nonce, user.password))
        assert crypto.encrypt(key, msg) == enc_token, 'enc_token'
        session = None if authorize.allow_multiple_sessions_per_user \
                  else db.fetch(Session, uid = user._uuid, fail = None)
        assert not session or force or authorize.timed_out(session), 'active'
        if session is None:
            session = db.insert(Session, uid = user._uuid, data = dict(),
                                         token = gen_token(), **argd)
        else:
            session.token = gen_token()
            db.update(session)
    return crypto.encrypt(key, dict(sid = session._uuid,
                                    name = user.full_name,
                                    token = session.token))

@service(safe = False, _log_exc = False)
def safe_call(sid, enc_call, **params):
    with store() as db:
        session = db.get(Session, sid, fail = None)
        user = db.get(Profile, session.uid, fail = None) if session else None
        assert user and not authorize.timed_out(session), 'expired'
        token, session.token = session.token, gen_token()
        call, args, argd = crypto.decrypt(token, enc_call)
        if params: argd.update(params)
        service.state.update(user = user, data = session.data)
        answer = dict(data = service.consume(call, args, argd, False),
                      token = session.token)
        db.update(session)
        return crypto.encrypt(token, answer)

@service()
def logout(_log_exc = False):
    service.state.clear()
    return True

@service(safe = False, auto_parse = False)
def arguments(*args, **argd):
    return dict(args = args, argd = argd)

@service(safe = False, auto_parse = True)
def params(*args, **argd):
    return dict(args = args, argd = argd)

@static(root = tempdir(), force_download = True)
@authorize()
def download(*path):
    return joinpath(*path)

@service()
@authorize()
def upload(*args, **argd):
    return [dict(name = filename(fh),
                 size = filesize(savefile(fh, tempdir())))
            for fh in files(*args, **argd)]

@service()
@authorize()
def find(*args, **argd):
    with store() as db:
        return db.select(*args, **argd)

@service()
@authorize()
def add(*args, **argd):
    with store() as db:
        return db.insert(*args, **argd)

@service()
@authorize()
def store(*args, **argd):
    with store() as db:
        return db.update(*args, **argd)

@service()
@authorize()
def remove(*args, **argd):
    with store() as db:
        return db.delete(*args, **argd)



from contextlib import contextmanager
from weakref import WeakValueDictionary

from fanery.terms import (
    parse_term, hict, get_timestamp, is_record_dump,
    is_class, is_string, is_date, is_number, is_boolean,
    is_uuid, is_sequence, is_record, is_record_ref,
    is_lambda, is_dict, is_file,
)
from fanery.crypto import digest, gen_token
from fanery.librand import randuuid
from fanery import service

class store(object):

    class MultipleRecordsFound(AssertionError):
        pass

    class NoResult(AssertionError):
        pass

    class DuplicatedKey(AssertionError):
        pass

    class ValidationError(AssertionError):
        pass

    class VersioningMismatch(AssertionError):
        pass

    _default_field_meta_definition = dict(
        # default setter/getter
        fget = lambda term, rec: (True, term),
        fset = lambda term, rec, old: (True, term),
        finit = None, # default to fset
        fdel = lambda term, rec: True,
        # auto parse term on set by default
        auto_parse = True,
        # use me as 2nd arg during validation
        use_on_validate = False,
        # default constaints
        validators = [],
        not_null = False,
        is_write_only = False,
        is_read_only = False,
        is_required = False,
        is_unique = False,
        is_key = False,
        # callable for transient fields
        fcomp = None)

    class data:

        class Term(dict):

            __getattr__ = dict.__getitem__
            __setattr__ = dict.__setitem__

            def __init__(self, basetype, **argd):
                assert isinstance(basetype, str)
                argd.update(basetype = basetype)
                is_unique = argd.setdefault('is_unique', False)
                is_required = argd.setdefault('is_required', is_unique)
                argd.setdefault('not_null', is_required)
                argd['validators'] = tuple(argd.setdefault('validators', []))
                argd.setdefault('use_on_validate', False)
                dict.__init__(self, **argd)

            def __call__(self, **argd):
                term = self.__class__(**self) # clone myself
                term.validators += tuple(argd.pop('validators', []))
                is_unique = argd.setdefault('is_unique', term.is_unique)
                is_required = argd.setdefault('is_required', is_unique or term.is_required)
                argd.setdefault('not_null', is_required or term.not_null)
                term.update(argd)
                return term

        Link = Term('Link', use_on_validate = True,
                fget = lambda term, rec: (True, parse_term(term)),
                validators = [lambda term, meta: is_record(term, meta.model) or 'bad-type'])

        Links = Term('Links', use_on_validate = True,
                fget = lambda term, rec: (True, parse_term(term)),
                validators = [lambda terms, meta: all(is_record(t, meta.model) for t in terms) or 'bad-type'])

        Boolean = Term('Boolean', validators = [is_boolean])

        Number = Term('Number', validators = [is_number])

        UUID = Term('UUID', validators = [is_uuid])

        Date = Term('Date', validators = [is_date])

        Dict = Term('Dict', validators = [is_dict])

        File = Term('File', validators = [is_file])

        List = Term('List', validators = [is_sequence])

        String = Term('String', validators = [is_string])

        Timestamp = Term('Timestamp', read_only = True, is_required = True,
                default = get_timestamp, validators = [is_number])

        NotEmptyString = String(not_null = True, validators = [lambda term: bool(term.strip()) or 'empty'])

        NotEmptyAlpha = NotEmptyString(validators = [lambda term: term.isalnum()])

        UniqueString = NotEmptyString(is_unique = True)

        UniqueAlpha = NotEmptyAlpha(is_unique = True)

        Password = NotEmptyString(is_required = True,
            fset = lambda term, rec, _: (True, digest(term)),
            validators = [lambda term: not term.isalnum() or 'too-simple',
                          lambda term: len(term) > 8 or 'too-short'])

    class Record(object):

        def __str__(self):
            return '<Record %s uuid: %s vsn: %s>' % (
                self._model, self._uuid, self._vsn)
        __repr__ = __str__

        def __getattr__(self, attr):
            if attr[0] == '_':
                raise AttributeError("'%s' record has no attribute '%s'" % (
                    self._model, attr))
            meta = self._meta_[attr]
            if not meta:
                meta._fill(store._default_field_meta_definition)
            elif meta.write_only:
                raise store.ValidationError('write-only', attr)
            elif meta.fcomp:
                return meta.fcomp(self)
            fields = self._fields
            if attr in fields:
                doit, term = meta.fget(fields[attr], self)
                if not doit:
                    raise store.ValidationError('write-only', attr)
                return term
            elif 'default' in meta:
                default = meta.default
                term = default() if callable(default) else default
                return fields.setdefault(attr, term)
            else:
                raise AttributeError("'%s' record attribute '%s' is unset %r" % (
                    self._model, attr, meta))

        def __setattr__(self, attr, term):
            if attr.startswith('_'):
                self.__dict__[attr] = term
            else:
                meta = self._meta_[attr]
                if not meta:
                    meta._fill(store._default_field_meta_definition)
                fields = self._fields
                if meta.read_only and attr in fields and fields[attr] != term:
                    raise store.ValidationError('read-only', attr)
                elif meta.fcomp:
                    raise store.ValidationError('transient', attr)
                elif meta.not_null and term is None:
                    raise store.ValidationError('not-null', attr)
                elif term is not None:
                    if meta.auto_parse:
                        term = parse_term(term)
                    for validator in meta.validators:
                        if meta.use_on_validate:
                            match = validator(term, meta)
                        else:
                            match = validator(term)
                        if match is not True:
                            raise store.ValidationError(match or (
                                'invalid' if is_lambda(validator) \
                                        else validator.__name__.replace('_','-')),
                                attr)
                if attr in fields:
                    doit, term = meta.fset(term, self, fields[attr])
                elif not callable(meta.finit):
                    doit, term = meta.fset(term, self, None)
                else:
                    doit, term = meta.finit(term, self)
                if not doit:
                    raise store.ValidationError('read-only', attr)
                fields[attr] = term

        def __delattr__(self, attr):
            meta = self._meta_[attr]
            if meta.fcomp:
                raise store.ValidationError('transient', attr)
            if meta.fdel(attr, self) is True:
                del self._fields[attr]

        def dump(self):
            return hict((k,v) for k,v in self.__dict__.items()
                        if k[0] == '_' and not k[-1] == '_')

        def load(self, dump):
            self.__dict__.update((k,v) for k,v in dump.items()
                    if k[0] == '_' and not k[-1] == '_')
            return self

    def __init__(self):
        self.models = dict()
        self.cache = dict()
        self.meta = hict()
        self.dbms = set()

    def set_definition(self, model, **argd):
        # extract/update fields definitions
        fields = dict((k, hict()._fill(v).merge(self._default_field_meta_definition))
                      for k,v in argd.items()
                      if isinstance(v, store.data.Term))
        # update model meta definition
        self.meta[model].merge(validators = [])\
            ._fill((k, argd[k]) for k in argd if k not in fields)\
            ._fill(unique = tuple(k for k in fields if fields[k].is_unique),
                   key = tuple(k for k in fields if fields[k].is_key),
                   fields = hict()._fill(fields))

    def join(self, dbm, *models):
        self.dbms.add(dbm)
        for model in models:
            self.models[model] = dbm.model(model)
            self.meta[model].merge(validators = [])
            self.cache[model] = WeakValueDictionary()

    @contextmanager
    def __call__(self):
        state = service.state
        if state.txn:
            yield self
        else:
            txn = state.txn = gen_token()
            dbms = self.dbms
            try:
                for dbm in dbms:
                    dbm.begin(txn)
                yield self
                for dbm in dbms:
                    dbm.prepare_commit(txn)
                for dbm in dbms:
                    dbm.commit(txn)
            except:
                for dbm in dbms:
                    dbm.rollback(txn)
                raise
            finally:
                state.txn = None

    def validate(self, record):
        meta = self.meta[record._model]
        # fields validators
        for field, f_meta in meta.fields.items():
            if f_meta.is_required \
                    and field not in record._fields \
                    and 'default' not in f_meta:
                raise store.ValidationError('required', field)
            try:
                fval = getattr(record, field)
            except:
                continue # unset but not required
            if fval is not None:
                for validator in f_meta.validators:
                    if f_meta.use_on_validate:
                        match = validator(fval, f_meta)
                    else:
                        match = validator(fval)
                    if match is not True:
                        raise store.ValidationError(match or (
                            'invalid' if is_lambda(validator) \
                                    else validator.__name__.replace('_','-')),
                            field, fval)
            elif f_meta.not_null:
                raise store.ValidationError('not-null', field)
        # record validators
        for validator in meta.validators:
            validator(record)
        # then check fields uniqueness
        fetch = self.models[record._model].fetch
        if meta.unique:
            for field in meta.unique:
                try:
                    fval = getattr(record, field)
                except:
                    continue # don't care for uniqueness if unset
                query = {field: getattr(record, field)}
                stored = fetch(**query)
                if stored and stored._uuid != record._uuid:
                    raise store.DuplicatedKey(query)
        # and finally check composite key uniqueness
        if meta.key:
            query = dict((field, getattr(record, field, None)) for field in meta.key)
            stored = fetch(**query)
            if stored and stored._uuid != record._uuid:
                raise store.DuplicatedKey(query)

    def new(self, model, *meta, **fields):
        record = store.Record()
        lmeta = len(meta)
        record.__dict__.update(
            _meta_ = self.meta[model].fields,
            _model = model,
            _uuid = meta[0] if lmeta else randuuid(),
            _txn = meta[1] if lmeta > 1 else None,
            _vsn = meta[2] if lmeta > 2 else 0,
            _time = meta[3] if lmeta > 3 else 0,
            _fields = meta[4] if lmeta > 4 else fields)
        return self.cache[model].setdefault((record._uuid, record._vsn), record)

    def _ret(self, ret, fail, model, *args):
        if ret:
            return self.cache[model].setdefault((ret._uuid, ret._vsn), ret)
        elif is_class(fail) and issubclass(fail, Exception):
            raise fail(model, *args)
        elif callable(fail):
            return fail()
        else:
            return fail

    def get(self, model, uuid, fail = NoResult):
        return self._ret(self.models[model].get(uuid), fail, model, uuid)

    def fetch(self, model, fail = NoResult, **argd):
        return self._ret(self.models[model].fetch(**argd), fail, model, argd)

    def select(self, model, **argd):
        cache = self.cache[model]
        for rec in self.models[model].select(**argd):
            yield cache.setdefault((rec._uuid, rec._vsn), rec)

    def insert(self, record_or_model, **argd):
        record = record_or_model if isinstance(record_or_model, store.Record) \
                 else self.new(record_or_model)
        for field, value in argd.items():
            if not field.startswith('_'):
                setattr(record, field, value)
        self.validate(record)
        record._vsn = 1
        record._txn = service.state.txn
        record._time = get_timestamp()
        self.models[record._model].insert(record)
        return self.cache[record._model].setdefault((record._uuid, 1), record)

    def update(self, record):
        assert record is self.cache[record._model][(record._uuid, record._vsn)]
        self.validate(record)
        record._vsn += 1
        record._txn = service.state.txn
        record._time = get_timestamp()
        self.models[record._model].update(record)
        return record

    def delete(self, record):
        key = (record._uuid, record._vsn)
        model = record._model
        cache = self.cache[model]
        assert record is cache[key]
        record._vsn *= -1
        record._txn = service.state.txn
        record._time = get_timestamp()
        self.models[model].delete(record)
        del cache[key]

    def clear(self, model):
        self.models[model].clear()
        self.cache[model].clear()

    def clearAll(self):
        cache = self.cache
        models = self.models
        for model in models:
            models[model].clear()
            cache[model].clear()



try:
    profile
except NameError:
    def profile(func):
        try:
            from cProfile import Profile
        except ImportError:
            from profile import Profile

        import functools

        @functools.wraps(func)
        def wrapper(*args, **argd):
            prof = Profile()
            try:
                return prof.runcall(func, *args, **argd)
            finally:
                prof.print_stats()

        return wrapper



from fanery import (
        service, authorize, store, Profile, BaseAcl, Admin,
        add_profile, check_permission, has_permission,
        grant_permission, fetch_record, select_records,
        insert_record, update_record, delete_record
)
from fanery.librand import randpasswd
from fanery.clients import PyClient, WSGIClient
from fanery.crypto import digest

@service(safe = False, ssl = False)
def anonymous():
    return True

@service(ssl = False)
@authorize()
def authenticated():
    return True

@service(ssl = False)
@authorize('admin')
def admin_only():
    return True

@service(ssl = False)
@authorize('user')
def user_only():
    return True

def init_db():
    with store() as db:
        db.insert(Profile, login = 'admin', role = 'admin', password = 'my-secret')
        db.insert(Profile, login = 'user', role = 'user', password = 'my-secret')

def test_authorize():
    init_db()

    for Client in (PyClient, WSGIClient):
        client = Client()
        assert client.call('anonymous') == True
        try:
            client.safe_call('authenticated')
            raise Exception('should have raised authorize.Unauthorized')
        except authorize.Unauthorized:
            client.login('user', 'my-secret')
            assert client.safe_call('authenticated') == True
            assert client.call('anonymous') == True
            client.logout()

        try:
            client.safe_call('admin_only')
            raise Exception('should have raised authorize.Unauthorized')
        except authorize.Unauthorized:
            pass

        client.login('user', 'my-secret')
        try:
            client.safe_call('admin_only')
            raise Exception('should have raised authorize.Unauthorized')
        except authorize.Unauthorized:
            assert client.call('anonymous') == True
            client.logout()
            client.login('admin', 'my-secret')
            assert client.safe_call('admin_only') == True

        try:
            client.safe_call('user_only')
            raise Exception('should have raised authorize.Unauthorized')
        except authorize.Unauthorized:
            assert client.call('anonymous') == True
            client.logout()
            client.login('user', 'my-secret')
            assert client.safe_call('user_only') == True
            client.logout()

def test_add_profile():
    with store() as db:
        db.clearAll()
    user = Admin
    name = 'Administrator'
    passwd = randpasswd()
    admin = add_profile(user, passwd, name = name, is_admin = True)
    assert admin.login == user
    assert admin.role == Admin
    assert admin.name == name
    assert admin.password == digest(passwd)
    # no permission for unauthenticated user
    for term in (admin, Profile, BaseAcl):
        assert not has_permission('select', term)
        assert not has_permission('insert', term)
        assert not has_permission('update', term)
        assert not has_permission('delete', term)
    # authenticated user
    service.state.update(user = admin, role = admin.role)
    assert has_permission('select', admin)
    assert has_permission('update', admin)
    assert not has_permission('insert', admin)
    assert not has_permission('delete', admin)
    for model in (Profile, BaseAcl):
        assert has_permission('select', model)
        assert has_permission('update', model)
        assert has_permission('insert', model)
        assert has_permission('delete', model)
    return admin

if __name__ == '__main__':
    test_authorize()



from fanery import store, service, Profile

def test_single_txn():
    session = service.state
    with store() as db:
        state = service.state
        assert state is session, (session, state)
        assert state.txn, state
        with store() as db1:
            assert service.state is state
            assert service.state.txn is state.txn, (service.state, state)

def test_data_terms():
    fields = ((store.data.Term('term'), False, False, False, 0),
              (store.data.Link(model = Profile), False, False, False, 1),
              (store.data.Links(model = Profile), False, False, False, 1),
              (store.data.Boolean(), False, False, False, 1),
              (store.data.Number(), False, False, False, 1),
              (store.data.UUID(), False, False, False, 1),
              (store.data.Date(), False, False, False, 1),
              (store.data.Dict(), False, False, False, 1),
              (store.data.File(), False, False, False, 1),
              (store.data.List(), False, False, False, 1),
              (store.data.String(), False, False, False, 1),
              (store.data.Timestamp(), False, True, True, 1),
              (store.data.NotEmptyString(), False, False, True, 2),
              (store.data.NotEmptyAlpha(name = 'nealpha'), False, False, True, 3),
              (store.data.UniqueString(name = 'ustr'), True, True, True, 2),
              (store.data.UniqueAlpha(name = 'ualpha'), True, True, True, 3),
              (store.data.Password(), False, True, True, 4))
    for field, unique, required, not_null, vlen in fields:
        name = field.get('name', field.basetype)
        assert isinstance(name, basestring), name
        assert field.is_unique is unique, (name, field, unique)
        assert field.is_required is required, (name, field, required)
        assert field.not_null is not_null, (name, field, not_null)
        assert isinstance(field.validators, tuple), (name, field)
        assert len(field.validators) == vlen, (name, field.validators, vlen)




from fanery.terms import parse_term
from fanery import store

class Model(object):

    def __init__(self, name, db):
        self.name = name
        self.records = db.records.setdefault(name, dict())

    def get(self, uuid):
        return self.records.get(uuid, None)

    def fetch(self, **argd):
        record = None
        for idx, record in enumerate(self.select(**argd)):
            if idx > 0:
                raise store.MultipleRecordsFound()
        return record

    def select(self, **argd):
        query = parse_term(argd).items()
        for record in self.records.values():
            fields = record.dump()._fields
            for k, v in query:
                if k not in fields or fields[k] != v:
                    break
            else:
                yield record

    def store(self, record):
        self.records[record._uuid] = record
    insert = update = store

    def delete(self, record):
        del self.records[record._uuid]

    def clear(self):
        self.records.clear()

class SimpleMemoryDB(object):

    def __init__(self):
        self.records = dict()

    def begin(self, txn):
        pass

    def prepare_commit(self, txn):
        pass

    def commit(self, txn):
        pass

    def rollback(self, txn):
        pass

    def model(self, name):
        return Model(name, self)



# -*- coding: utf-8 -*-

from dateutil import tz
from dateutil.rrule import MO, TU, WE, TH, FR, SA, SU

ISO2 = 'CO'
TZ = tz.gettz('America/Bogota')
WORKDAY = (MO, TU, WE, TH, FR)
WEEKEND = (SA, SU)

def holidays(year = None):
    # from https://es.wikipedia.org/wiki/Anexo:Días_feriados_en_Colombia
    from fanery.locales import easter, relativedelta as rd
    from dateutil.rrule import rrule, WEEKLY, MO
    from datetime import datetime as dt
    year = year or dt.today().year
    easter = easter(year)
    # first monday
    fm = rd(weekday = MO(+1))
    return (
            dt(year, 1, 1),         # Año Nuevo
            dt(year, 1, 6) + fm,    # Epifanía (Reyes Magos)
            dt(year, 3, 19) + fm,   # San José
            easter + rd(days = -7), # Domingo de Ramos
            easter + rd(days = -3), # Jueves Santo
            easter + rd(days = -2), # Viernes Santo
            easter,                 # Domingo de Pascua
            dt(year, 5, 1),         # Día del Trabajo
            easter + rd(days = 43), # Ascensión de Jesús
            easter + rd(days = 64), # Corpus Christi
            easter + rd(days = 71), # Sagrado Corazón de Jesús
            dt(year, 6, 29) + fm,   # san Pedro y san Pablo
            dt(year, 7, 20),        # Grito de Independencia
            dt(year, 8, 7),         # Batalla de Boyacá
            dt(year, 8, 15) + fm,   # Asunción de la Virgen
            dt(year, 10, 12) + fm,  # Día de la Raza
            dt(year, 11, 1) + fm,   # Todos los Santos
            dt(year, 11, 11) + fm,  # Independencia de Cartagena[
            dt(year, 12, 8),        # Inmaculada Concepción
            dt(year, 12, 25))       # Navidad



from functools import wraps

from fanery.crypto import gen_token, digest, encrypt, decrypt
from fanery.terms import (
    parse_term, is_string, is_sequence,
    is_subclass, hict, get_timestamp,
)
from fanery import service, store

# Default authorization models
Profile = 'Profile'
Session = 'Session'
BaseAcl = 'BaseAcl'

# Default roles
Admin = 'admin'
User = 'user'

# Default storage for authorization models (testing purpose only)
from fanery.dbms.SimpleMemoryDB import SimpleMemoryDB
store.join(SimpleMemoryDB(), Profile, Session, BaseAcl)

# Profile constraints
store.set_definition(Profile,
    role        = store.data.NotEmptyAlpha(is_required = True),
    login       = store.data.UniqueAlpha(),
    password    = store.data.Password())

# BaseAcl constraints
store.set_definition(BaseAcl,
    rtype   = store.data.NotEmptyAlpha(is_key = True, is_required = True,
                validators = [lambda term: term in store.models or 'unknown-model']),
    rid     = store.data.UUID(is_key = True))

class authorize(object):
    '''Default session authorization implementation.'''

    # in seconds (2 hours)
    SESSION_TIMEOUT = 7200

    # define single session check
    ALLOW_MULTIPLE_SESSIONS_PER_USER = False

    class Unauthorized(AssertionError):
        pass

    class SessionActive(AssertionError):
        pass

    class SessionTimeout(AssertionError):
        pass

    class InvalidCredentials(AssertionError):
        pass

    def __init__(self):
        self.validators = dict(
            users = lambda users: service.state.user.login in users,
            roles = lambda roles: service.state.role in roles,
            grant = lambda grant: check_permission('call', grant))

    def __call__(self, *args, **argd):
        '''Authorization decorator.

        Must be used before @service decorator, like this:

            @service()
            @authorize()
            def auth_only(*args, **argd):
                """Safe and auth protected service.

                Encrypted service call allowed only to authenticated users.
                """
                do something
        '''
        params = hict((k, set(v) if is_sequence(v) else v)
                      for k, v in argd.items())
        if args and not params.roles:
            params.roles = set(args)

        validators = self.validators

        def decorator(func):

            func_name = '%s.%s' % (func.__module__, func.__name__)

            @wraps(func)
            def wrapper(*args, **argd):
                user = service.state.user
                if user is None:
                    raise self.Unauthorized(func_name, user)
                for k, v in params.items():
                    if k in validators and validators[k](v) is False:
                        raise self.Unauthorized(func_name, (user.login, user.role), (k, v))
                return func(*args, **argd)

            return wrapper

        return decorator

    def add_validator(self, name, validator):
        '''Plug new authorization validator.

        name      - the decorator argument associated with validator.
        validator - callable object (must return False to raise Unauthorized).
        '''
        assert callable(validator)
        self.validators[name] = validator

    @staticmethod
    def timed_out(session):
        '''Return True if session timed out (see authorize.SESSION_TIMEOUT).'''
        return (session.tstamp + authorize.SESSION_TIMEOUT) < get_timestamp()

@service(ssl = False, safe = False, auto_parse = False)
def login(username, nonce, tstamp, enc_token, force = False, **argd):
    '''Encrypted user login (default session authorization).'''
    with store() as db:
        user = db.fetch(Profile, login = username, fail = None,
                        **(parse_term(argd) if argd else argd))
        if user is None:
            raise authorize.InvalidCredentials('username')
        try:
            key = digest('%s%s%f' % (nonce, user.password, parse_term(tstamp)))
        except:
            raise AssertionError()
        msg = digest('%s%s%s' % (user.login, nonce, user.password))
        if encrypt(key, msg) != enc_token:
            raise authorize.InvalidCredentials('password',
                    '%s%s%s' % (user.login, nonce, user.password),
                    key, msg, user.login, argd)
        session = None if authorize.ALLOW_MULTIPLE_SESSIONS_PER_USER \
                  else db.fetch(Session, uid = user._uuid, fail = None)
        if session and not (force is True or authorize.timed_out(session)): \
            raise authorize.SessionActive()
        elif session is None:
            session = db.insert(Session, uid = user._uuid, role = user.role,
                                token = gen_token(), tstamp = get_timestamp(),
                                data = dict(), **argd)
        else:
            session.tstamp = get_timestamp()
            session.token = gen_token()
            db.update(session)
    return encrypt(key, dict(sid = session._uuid, token = session.token,
                             user = dict((k,v) for k,v in user._fields.items()
                                         if k[0] != '_' and k != 'password')))

@service(log_exc = True, ssl = False, safe = False, auto_parse = False)
def safe_call(sid, enc_call, **params):
    '''Encrypted call (using default session authorization).'''
    with store() as db:
        session = db.get(Session, parse_term(sid), fail = None)
        if session is None:
            raise authorize.Unauthorized('session')
        elif authorize.timed_out(session):
            raise authorize.SessionTimeout()
        user = db.get(Profile, session.uid, fail = None) if session else None
        if user is None:
            raise authorize.Unauthorized('user')
        token = session.token
        try:
            call, args, argd = decrypt(token, enc_call)
        except:
            raise authorize.Unauthorized('token')
        state = service.state.update(safe = True, user = user,
                        role = session.role, data = session.data)
        if params:
            argd.update(params)
        # consume service and update session token unless static
        call, _, _, ret = service._consume(call, *args, **argd)
        if call.static is True:
            return ret
        elif state.user:
            if call.one_time_pad is True:
                session.token = new_token = gen_token()
            else:
                session.token = new_token = token
            session.tstamp = get_timestamp()
            db.update(session)
        else:
            db.delete(session)
            new_token = None
        return encrypt(token, dict(data = ret, token = new_token))

@service(ssl = False, auto_parse = False)
def logout():
    '''Secure logout.'''
    service.state.clear()
    return True

def grant_permission(*terms, **rules):
    disable = rules.pop('disable', False)
    replace = rules.pop('replace', False)
    merge = rules.pop('merge', False)
    rules = hict(rules)
    for action, action_rules in rules.items():
        if isinstance(action_rules, dict):
            rules[action] = hict((getattr(t, '_uuid', t), bool(b))
                                 for t,b in action_rules.items())
        elif isinstance(action_rules, bool):
            rules[action] = {'*': action_rules}
        else:
            raise TypeError('dict or bool rules expected: %r' % action_rules)
    with store() as db:
        for term in terms:
            if isinstance(term, store.Record):
                rtype = term._model
                rid = term._uuid
            elif is_string(term):
                rtype = term
                rid = None
            else:
                raise TypeError('Record or model name expected: %r' % term)
            acl = db.fetch(BaseAcl, rtype = rtype, rid = rid, fail = None)
            if acl:
                check_permission('update', acl)
                if replace:
                    acl.rules.update(rules)
                else:
                    for action, action_rules in rules.items():
                        actual_rules = acl.rules[action]
                        if merge:
                            actual_rules.merge(action_rules)
                        else:
                            actual_rules.update(action_rules)
                acl.disabled = disable
                db.update(acl)
            else:
                user = service.state.user
                if user:
                    check_permission('insert', BaseAcl)
                    allow = {Admin: True, user._uuid: True}
                else:
                    allow = {Admin: True}
                acl = db.insert(BaseAcl, rtype = rtype, rid = rid,
                                rules = rules, disabled = disable)
                db.insert(BaseAcl, rtype = BaseAcl, rid = acl._uuid, disabled = False,
                                   rules = dict(select = allow,
                                                insert = {'*': False},
                                                update = allow,
                                                delete = allow))

def check_permission(action, term, uuid = None, fail = authorize.Unauthorized):
    state = service.state
    role = state.get('role', None)
    user = getattr(state.get('user', None), '_uuid', None)
    if isinstance(term, store.Record):
        term, uuid = term._model, term._uuid
    if action == 'insert' and uuid:
        rules = False
    else:
        with store() as db:
            acl = db.fetch(BaseAcl, rtype = term, rid = uuid, fail = None)
            rules = acl.rules.get(action, None) if acl and not acl.disabled else None
            if uuid and not (rules is False or (rules and (role in rules or user in rules))):
                acl = db.fetch(BaseAcl, rtype = term, rid = None, fail = None)
                rules = (role == Admin) if not acl else (acl.rules.get(action, False) \
                                                         if not acl.disabled else False)
    if not (rules and rules.get(user, rules.get(role, rules.get('*', False)))):
        if is_subclass(fail, Exception):
            raise fail(role, action, term, uuid, role, rules)
        else:
            return fail
    else:
        return True

def has_permission(*args, **argd):
    argd.setdefault('fail', False)
    return check_permission(*args, **argd)

def add_profile(login, password, role = None, is_admin = False, **extra):
    role = role or (Admin if is_admin else User)
    is_admin = bool(role == Admin)
    with store() as db:
        grant_permission(BaseAcl, Profile, select = {role: True},
                                           insert = {role: True},
                                           update = {role: is_admin},
                                           delete = {role: is_admin})
        user = db.insert(Profile, role = role, login = login,
                                  password = password, **extra)
        grant_permission(user, select = {user: True},
                               insert = False,
                               update = {user: True},
                               delete = {user: False})
        return user

def fetch_record(model, *args, **argd):
    with store() as db:
        record = db.get(model, *args) if args else db.fetch(model, **argd)
        check_permission('select', record)
        return record

def select_records(model, **argd):
    check_permission('select', model)
    with store() as db:
        for record in db.select(model, **argd):
            if has_permission('select', record):
                yield record

def insert_record(model, **data):
    check_permission('insert', model)
    with store() as db:
        return db.insert(model, **data)

def update_record(model, uuid, vsn, **data):
    check_permission('update', model, uuid)
    with store() as db:
        record = db.get(model, uuid)
        if vsn != record._vsn:
            raise store.VersioningMismatch()
        for field, value in data.items():
            setattr(record, field, value)
        return db.update(record)

def delete_record(model, uuid):
    check_permission('delete', model, uuid)
    with store() as db:
        record = db.get(model, uuid)
        db.delete(record)
        return True



from random import randint, randrange, sample, shuffle, choice
from random import getrandbits as randbits
from random import uniform, random
from os import urandom as randbytes
from uuid import uuid4 as randuuid

def randpasswd(size = 12):
    if not (isinstance(size, (int, long)) and size > 0):
        raise ValueError('size must be positive number')
    from string import printable
    chars = list(set(printable.strip()))
    shuffle(chars)
    return ''.join(sample(chars, size))

def randfloat(*args):
    return uniform(*args) if args else random()

def randnum(*args, **argd):
    exp = argd.get('exp', 300)
    sign = choice([1, -1])
    try:
        return sign * pow(randfloat(*args), randint(-exp,exp))
    except OverflowError:
        return randnum(*args, **argd)



__all__ = ['gen_token', 'digest', 'xor', 'encrypt', 'decrypt']

from numpy import frombuffer, bitwise_xor, byte
from base64 import b64encode, b64decode
from hashlib import sha256 as hashfun
from os import urandom as rand_bytes
from fanery.terms import is_string, to_json, parse_json

def gen_token(size = 32):
    return b64encode(rand_bytes(size))

def digest(term):
    return b64encode(hashfun(term if is_string(term) else str(term)).digest())

def xor(key, msg):
    key_buff = frombuffer(key, dtype = byte)
    msg_buff = frombuffer(msg, dtype = byte)
    return bitwise_xor(key_buff, msg_buff).tostring()

def cipher(key, msg):
    k = digest(key)
    klen = len(k)
    div, mod = divmod(len(msg), klen)
    for i in xrange(div):
        k = digest(k)
        yield xor(k, msg[i*klen:(i+1)*klen])
    if mod:
        k = digest(k)
        yield xor(k[:mod], msg[-mod:])

def encrypt(key, term):
    return b64encode(''.join(cipher(key, to_json(term))))

def decrypt(key, enc):
    return parse_json(''.join(cipher(key, b64decode(enc))))



from setuptools import setup

setup(
    name = 'Fanery',
    version = '0.1.1',
    author = 'Marco Caramma',
    author_email = 'marco@globalsoftwaresecurity.com',
    packages = [
        'fanery',
        'fanery.dbms',
        'fanery.locales',
        'fanery.test',
        ],
    url = 'http://pypi.python.org/pypi/Fanery/',
    license = 'BSD',
    description = 'Application development framework',
    long_description = open('README.txt').read(),
    install_requires = [
        'python-dateutil',
        #'memory_profiler',
        #'line_profiler',
        #'pprofiler',
        #'objgraph',
        'psutil',
        'numpy',
        'webob',
        #'nose',
        #'rawes',
        #'netaddr',
        #'pygeoip',
        #'pycountry',
        ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from fanery import service

def html(term):
    return '<p>%s</p>' % term

def xml(term):
    return '<%(tag)s>%(value)s</%(tag)s>' % \
        dict(value = term, tag = type(term).__name__)

@service(html = html, xml = xml)
def safe_echo(term = None):
    return term

@service(safe = False, ssl = False, html = html, xml = xml)
def echo(term = None):
    return term

def _state():
    return service.state

@service(safe = False, ssl = False)
def state1():
    return service.state, _state()

@service(safe = False, ssl = False)
def state2():
    return service.state, _state()

def test_service_meta():
    assert safe_echo.service.safe == True
    assert safe_echo.service.ssl == True
    assert echo.service.route == echo.__name__
    assert echo.service.cache == False
    assert echo.service.safe == False
    assert echo.service.ssl == False
    assert callable(echo.service.output.xml)
    assert callable(echo.service.output.html)
    assert echo.service.output.xml is xml
    assert echo.service.output.html is html
    assert echo.service.output == safe_echo.service.output

def test_lookup():
    func, args, exc, output = service.lookup('echo.xls')
    assert func is echo and not args and exc == '.xls' and output is None
    try:
        service.lookup('not-registered')
        raise Exception('should raise service.NotFound')
    except service.NotFound:
        pass

def test_consume():
    assert service.consume('echo') is None
    assert service.consume('echo/1') is 1
    assert service.consume('echo/1.0') is 1
    assert service.consume('echo', 1) is 1
    assert service.consume('echo', '1') is 1
    assert service.consume('echo', 1.0) is 1
    assert service.consume('echo', '1.0') is 1
    assert service.consume('echo', term = 1) is 1
    assert service.consume('echo', term = '1') is 1
    assert service.consume('echo', term = 1.0) is 1
    assert service.consume('echo', term = '1.0') is 1

def test_output():
    html = '<p>1</p>'
    assert service.consume('echo/1.html') == html
    assert service.consume('echo/1.0.html') == html
    assert service.consume('echo.html', 1) == html
    assert service.consume('echo.html', '1') == html
    assert service.consume('echo.html', 1.0) == html
    assert service.consume('echo.html', '1.0') == html
    assert service.consume('echo.html', term = 1) == html
    assert service.consume('echo.html', term = '1') == html
    assert service.consume('echo.html', term = 1.0) == html
    assert service.consume('echo.html', term = '1.0') == html

    xml = '<int>1</int>'
    assert service.consume('echo/1.xml') == xml
    assert service.consume('echo/1.0.xml') == xml
    assert service.consume('echo.xml', 1) == xml
    assert service.consume('echo.xml', '1') == xml
    assert service.consume('echo.xml', 1.0) == xml
    assert service.consume('echo.xml', '1.0') == xml
    assert service.consume('echo.xml', term = 1) == xml
    assert service.consume('echo.xml', term = '1') == xml
    assert service.consume('echo.xml', term = 1.0) == xml
    assert service.consume('echo.xml', term = '1.0') == xml

    xml = '<date>2012-01-30</date>'
    assert service.consume('echo/2012-1-30.xml') == xml
    assert service.consume('echo.xml', '30 Jan 2012') == xml
    assert service.consume('echo.xml', term = '01/30/2012') == xml

def test_state():
    session1_1, session1_2 = state1()
    session2_1, session2_2 = state2()
    assert session1_1 is session1_2
    assert session2_1 is session2_2
    assert session1_1 is not session2_1



from mimetypes import types_map
from webob import dec, exc, static, Response

from fanery import service, authorize
from fanery.terms import (
    hict, is_file, is_dir, is_string,
    _from_json as parse_json, to_json,
)

SERVER_NAME = 'FWS/1.0'
CONTENT_TYPE = 'text/plain'
JSON_KEY = '_json'

@dec.wsgify
def handler(req):
    # read request params
    try:
        params = req.params
        getall = params.getall
        argd = dict((k, (v[0] if len(v) == 1 else v) or None) for k, v in [
                     (k, [v.strip() for v in getall(k)]) for k in params
                                                  if not k.startswith('_')])
        if JSON_KEY in params:
            argd.update(parse_json(params[JSON_KEY]))
    except Exception, e:
        raise exc.HTTPBadRequest(*e.args)

    # start session state
    service.state.update(ssl = bool(req.scheme == 'https'),
                         domain = req.host.split(':')[0],
                         origin = req.remote_addr)

    # consume remote call
    try:
        call, ext, output, ret = service._consume(req.path, **argd)
        if call.static is True:
            if is_dir(ret):
                return static.DirectoryApp(ret)
            elif is_file(ret):
                if isinstance(ret, file):
                    with ret:
                        ret = ret.name
                if call.force_download is True:
                    return static.FileApp(ret,
                        content_type = "application/force-download",
                        content_disposition = 'attachment; filename="%s"' % ret)
                else:
                    return static.FileApp(ret)
            else:
                raise service.NotFound(ret)
        else:
            res = Response(content_type = types_map.get(ext, CONTENT_TYPE))
            res.body_file.write(ret if is_string(ret) else to_json(ret))
            res.cache_expires(call.cache)
            res.server = SERVER_NAME
            res.md5_etag()
            return res
    except service.NotFound, e:
        raise exc.HTTPNotFound()
    except (NotImplementedError, service.UnknownFormatter), e:
        raise exc.HTTPNotImplemented(*e.args)
    except AssertionError, e:
        res = Response(content_type = CONTENT_TYPE)
        res.status_code = 400 # bad request
        error = dict(type = e.__class__.__name__, reason = e.args)
        res.body_file.write(to_json(error))
        return res
    except Exception, e:
        raise exc.HTTPInternalServerError(*e.args)

'''
try:
    from gevent import ...
except ImportError:
    try:
        from cherrypy.wsgiserver import CherryPyWSGIServer
    except ImportError:
        from wsgiref import ....
'''



import terms
import filelib
import librand
import crypto

from _service import service
service = service()

from _store import store
terms.store = store = store()

from _authorize import (
        authorize, Profile, Session, BaseAcl, Admin, User,
        grant_permission, check_permission, has_permission,
        add_profile, fetch_record, select_records,
        insert_record, update_record, delete_record,
)
authorize = authorize()
